import json

from django.test import TestCase 
from django.contrib.sites.models import Site
from django.contrib.auth import get_user_model
from django.db.utils import IntegrityError
from django.db import transaction

from permafrost.models import Role, ROLE_CONFIG, get_permission_models

class RoleModelTest(TestCase):

    # fixtures = ['unit_test']

    def setUp(self):
        User = get_user_model()
        self.user = User.objects.create_user(username='john', email='jlennon@beatles.com', password='Passw0rd!')
        self.staffuser = User.objects.create_user(username='staffy', email='staffy@beatles.com', password='Passw0rd!')
        self.adminuser = User.objects.create_user(username='adminy', email='adminy@beatles.com', password='Passw0rd!')

        self.site_1 = Site.objects.create(name="This Site", domain="thissite.com")
        self.site_2 = Site.objects.create(name="That Site", domain="thatsite.com")

    #     self.client_user = get_user_model().objects.get(pk=5)
    #     self.client_staff = get_user_model().objects.get(pk=6)
    #     self.client_admin = get_user_model().objects.get(pk=7)

    def test_permission_objects_from_string(self):
        perms = get_permission_models("permafrost.view_role")

        self.assertEqual(1, len(perms))

    def test_permission_objects_from_string_list(self):
        perms = get_permission_models( ["permafrost.view_role", "permafrost.change_role"] ) 

        self.assertEqual(2, len(perms))

    def test_create_user_role(self):
        # Test that creating a Role creates a matching Group

        role = Role(name="Bobs Super Group")
        role.save()

        self.assertEqual(role.group.name, "1_user_bobs-super-group")        # Checks that the user is created

        # Add the user to the Role with the 
        role.users_add(self.user)

        # perms = [x.name for x in role.group.permissions.all()]
        # Get the list from the user?
        perms = list(self.user.get_all_permissions())

        # Test that "included" permissions are always present in the group
        self.assertEqual(perms, [])

    # def test_role_permission_limits(self):
        # Test that creating a Role creates a matching Group

        # Test that permissions only in the Category's list can be added to the Role

            # Get a list of the current permissions

            # try to add a permission

            # Check to see if it was added

    def test_create_staffuser_role(self):
        # Test that creating a Role creates a matching Group

        role = Role(name="Bobs Staff Group", category=30)
        role.save()

        self.assertEqual(role.group.name, "1_staff_bobs-staff-group")        # Checks that the user is created

        # Add user to the Group
        role.users_add(self.staffuser)
        perms = list(self.staffuser.get_all_permissions())
        perms.sort()

        check_list = ["permafrost.view_role",  "permafrost.change_role"]    # Needs to be sorted to try to make sure it's a close to the same as possible
        check_list.sort()

        # Test that "included" permissions are the only things present in the group
        self.assertEqual(perms, check_list)

    def test_only_allowed_category_perms_can_be_added(self):
        # Test that permissions not in the Category's list can not be added to the Role
        pass

    def test_create_duplicate_role(self):
        # Test that creating a Role of the same name producers and error
        role_a = Role(name="Bobs Super Group", site=self.site_1)
        role_a.save()

        role_c = Role(name="Bobs Super Group", site=self.site_2)
        role_c.save()

        with self.assertRaises(IntegrityError):

            with transaction.atomic():
                role_b = Role(name="Bobs Super Group", site=self.site_2)
                role_b.save()

            with transaction.atomic():
                role_d = Role(name="Bobs Super Group", site=self.site_2, category=30)
                role_d.save()

        # TODO: Add check to make sure Role's Groups were not created

    def test_clear_role_permissions(self):
        # Test that "included" permissions are always present in the group

        role = Role(name="Bobs Staff Group", site=self.site_2, category=30)

    #     Add Extra Permissions from the list
    #     Clear
    #     Make sure only the includes are present
    #     raise NotImplementedError

    # Test the Client Users permissions

    # def test_django_admin_has_all_permissions(self):
    #     # Test the Django Users permissions
    #     raise NotImplementedError

    # Test that deleting a Role deletes the matching group


#     def test_superuser_has_perms(self):
#         '''
#         Test that a user with
#         '''
#         View = self.get_view("any", ['edit-step', 'edit-path'])
#         request = self.get_request('/', self.superuser)

#         response = View.as_view()(request)
#         self.assertEqual(response.status_code, 200)



from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from django.contrib.auth.models import Group, Permission
from django.contrib.sites.shortcuts import get_current_site
from django.utils.translation import ugettext_lazy as _
from django.utils.text import slugify


'''
A ROLE_CATEGORY defines what permissions are assignable per category
These are defined by developers and used by clients.
Only what is available here will be available to the Client to set.
For security reasons, this should be hard coded.

Example:

PERMAFROST_ROLE_CONFIG = {
    'User': {                   # Permission Grouping
        'permissions': [        # List of Django Permissions that are client configurable (as string or dict)
                {"perm":"permafrost.view_role", "label":"Can view Role"},
                {"perm":"permafrost.view_rolepermission", "label":"Can view Role Permission"},
            ],
        'id': 1,             # The Value stored in the Choice Field
        'includes': [           # Permissions the Group uses that are always included but not editable by the client
            '',
            ''
        ]
    },
    'Staff': {
        'permissions': [        # List of Django Permissions that are client configurable (as string or dict)
                {"perm":"permafrost.add_role", "label":"Can add Role"},
                {"perm":"permafrost.change_role", "label":"Can change Role"},
                {"perm":"permafrost.view_role", "label":"Can view Role"},
                {"perm":"permafrost.add_rolepermission", "label":"Can add Role Permission"},
                {"perm":"permafrost.change_rolepermission", "label":"Can change Role Permission"},
                {"perm":"permafrost.view_rolepermission", "label":"Can view Role Permission"},
            ],
        'id': 30,            # The Value stored in the Choice Field
        'includes: [
            "permafrost.view_role",
            "permafrost.view_rolepermission",
        ]
    },
    'Administrator': {
        'permissions': [        # List of Django Permissions that are client configurable (as string or dict)
                {"perm":"permafrost.add_role", "label":"Can add Role"},
                {"perm":"permafrost.change_role", "label":"Can change Role"},
                {"perm":"permafrost.delete_role", "label":"Can delete Role"},
                {"perm":"permafrost.view_role", "label":"Can view Role"},
                {"perm":"permafrost.add_rolepermission", "label":"Can add Role Permission"},
                {"perm":"permafrost.change_rolepermission", "label":"Can change Role Permission"},
                {"perm":"permafrost.delete_rolepermission", "label":"Can delete Role Permission"},
                {"perm":"permafrost.view_rolepermission", "label":"Can view Role Permission"},
            ],
        'id': 50,
        'includes': [
            '',
            ''
        ]
    }
}
'''

ROLE_CONFIG = getattr(settings, "PERMAFROST_ROLE_CONFIG")

###############
# CHOICES
###############

ROLE_CATEGORY_CHOICES = [(value['id'], key) for key,value in ROLE_CONFIG.items()]      # Used to help manage permission presentation

###############
# UTILITIES
###############

def lable_from_choice_value(choices, value):
    return [x[1] for x in choices if x[0] == value][0]

def value_as_list(value):
    '''
    Takes whatever is passed in and tries to return it as a list
    '''
    if isinstance(value, list):
        return value
    
    if not value:
        return []

    return [value]

def get_permission_models(permissions):
    return [ permission_from_string(p) for p in value_as_list(permissions) ]

def permission_from_string(permission):
    values = permission.split(".")
    return Permission.objects.get(codename=values[1], content_type__app_label=values[0])


###############
# MODELS
###############

class Role(models.Model):
    '''
    Role is Client Defineable and "wraps" around a Django Group
    adding a user to this role adds them to the Django Group
    and automatically assignes them the permissions.
    '''
    name = models.CharField(_("Name"), max_length=50)
    slug = models.SlugField(_("Slug"))
    category = models.IntegerField(_("Role Category"), choices=ROLE_CATEGORY_CHOICES, default=1)
    site = models.ForeignKey(Site, on_delete=models.CASCADE, default=settings.SITE_ID)
    group = models.OneToOneField(Group, verbose_name=_("Group"), on_delete=models.CASCADE, blank=True, null=True)      # Need to be uneditable in the Admin
    locked = models.BooleanField(_("Locked"), default=False)                                                   # If this is locked, it can not be edited by the Client, used for defaults

    class Meta:
        verbose_name = _("Role")
        verbose_name_plural = _("Roles")
        unique_together = [['name', 'site']]

    def __str__(self):
        return self.name

    #-------------
    # Permissions

    def permissions_add(self, permissions):
        '''
        Add permissions to the attached group by name "app.perm"
        '''
        for p in get_permission_models(permissions):
            self.group.permissions.add(p)

    def permissions_remove(self, permissions):
        '''
        Remove permissions from the attached group by name "app.perm"
        '''
        for p in get_permission_models(permissions):
            self.group.permissions.remove(p)

    def permissions_set(self, permissions):
        '''
        This updates the group permissions to only include what was passed in.
        '''
        self.group.permissions.set( value_as_list(permissions) )

    def permissions_clear(self):
        '''
        Remove all permissions from the group except the defaults.
        '''
        category = lable_from_choice_value(ROLE_CATEGORY_CHOICES, self.category)

        if 'includes' in ROLE_CONFIG[category]:
            self.group.permissions.set( value_as_list(ROLE_CONFIG[category]['includes']) )
        else:
            self.group.permissions.clear()


    #-------------
    # Users

    def users_add(self, users=None):
        '''
        Pass in a User object to add to the Role
        '''
        for user in value_as_list(users):
            user.groups.add(self.group)

    def users_remove(self, users=None):
        '''
        Pass in a User object to remove from the Role
        '''
        for user in value_as_list(users):
            user.groups.remove(self.group)

    def users_clear(self):
        '''
        Remove all users from the Role
        '''
        self.group.clear()

    #-------------
    # Save

    def save(self, *args, **kwargs):

        self.slug = slugify(self.name)

        result = super().save(*args, **kwargs)

        if not self.group:
            category = lable_from_choice_value(ROLE_CATEGORY_CHOICES, self.category)
            obj, created = Group.objects.get_or_create( name="{0}_{1}_{2}".format( self.site.pk, slugify(category), slugify(self.name)) )

            if created:
                obj.save()                  # Save it to create the PK so it can be assigned
                self.group = obj
                perms = [ permission_from_string(permission) for permission in value_as_list(ROLE_CONFIG[category]['includes']) ]
                self.permissions_set( perms )    # If a new group is geenrated, the permissions shoudl be blank by default

            super().save(*args, **kwargs)   # Save again only if the Group is added

        return result



from django.contrib import admin

from .models import Role

# ################
# # ADMIN ACTIONS
# ################

# def recreate_slug(modeladmin, request, queryset):
#     '''
#     This uses the save() instead of update() since the AutoSlugField is only generated on save().
#     '''
#     for item in queryset.all():
#         item.slug = ""
#         item.save()

# recreate_slug.short_description = "Regenerate the Slug Field"


# def make_enabled(modeladmin, request, queryset):
#     queryset.update(enabled=True)

# make_enabled.short_description = "Mark Enabled"


# def make_disabled(modeladmin, request, queryset):
#     queryset.update(enabled=False)

# make_disabled.short_description = "Mark Disabled"


###############
# MODEL ADMINS
###############

class RoleAdmin(admin.ModelAdmin):
    # readonly_fields = ('slug',)
    # list_display = ('name', 'category')
    # ordering = ('name',)
    readonly_fields=('group','slug')


# class RolePermissionAdmin(admin.ModelAdmin):
#     # readonly_fields = ('slug',)
#     # list_display = ('name', 'category')
#     # ordering = ('name',)
#     pass


###############
# REGISTRATION
###############

admin.site.register(Role, RoleAdmin)
# admin.site.register(RolePermission, RolePermissionAdmin)



import logging

from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.views.generic import TemplateView

#--------------
# UTILITIES
#--------------

def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


#--------------
# MIXIN VIEWS
#--------------

class MethodPermissionRequiredMixin(PermissionRequiredMixin):
    '''
    This is a simple mixin that extend the built in PermissionRequiredMixin 
    and lets a developer specify perms required by a user for a particular 
    http method.  For example, a user with a 'get' pemission might not be
    given permissions to reach a 'post' endpoint in the view.
    
    If they don't have the required permissions, then the user
    is rejected.

    Permissions can be set per HTTP method, by appending the lowercase
    method name to 'permission_required_' and providing a set of permissions.

    permission_required = ('sites.add_site',)
    permission_required_get = ()
    permission_required_post = ()
    '''

    def get_permission_required(self):
        """
        Override this method to override the permission_required attribute.
        Must return an iterable.
        """
        perms = super().get_permission_required()

        method_perms = getattr(self, "permission_required_" + request.method.lower(), set() )       # Extended Perms per method

        if isinstance(method_perms, str):
            method_perms = (method_perms,)

        return set(perms + method_perms)


class LogPermissionRequiredMixin(object):
    '''
    A mixin that lets you define a logger in which to write failed permission attempts to.
    '''

    no_permission_logger = None

    def handle_no_permission(self):
        
        if self.no_permission_logger is None:
            raise ImproperlyConfigured(
                '{0} is missing the no_permission_logger attribute. Define {0}.no_permission_logger'.format(self.__class__.__name__)
            )

        logger = logging.getLogger(self.no_permission_logger)

        user_ip = get_client_ip(self.request)
        user_perms = list(self.request.user.get_all_permissions())
        view_perms = list(self.get_permission_required())

        logger.info("Failed-Permission-Check:403:{0}:{1}:{2}:{3}:{4}:{5}:{6}".format(                   # Should be replaced with a Formater
                                user_ip, self.request.user.username, self.request.user.pk, 
                                self.request.method, self.request.path, ','.join(user_perms),
                                ','.join(view_perms) )
                            )

        super().handle_no_permission()
